x86: allocate membarrier system call number
[urcu.git] / urcu.c
diff --git a/urcu.c b/urcu.c
index ae3490f7cb2cdc1f7cdb1d26f77c0ee22620e9eb..a620b762b4796712bd05efb862de8cb6c53c41fd 100644 (file)
--- a/urcu.c
+++ b/urcu.c
@@ -26,6 +26,7 @@
 #define _BSD_SOURCE
 #define _GNU_SOURCE
 #define _LGPL_SOURCE
+#define _DEFAULT_SOURCE
 #include <stdio.h>
 #include <pthread.h>
 #include <signal.h>
@@ -36,6 +37,7 @@
 #include <errno.h>
 #include <poll.h>
 
+#include "urcu/arch.h"
 #include "urcu/wfcqueue.h"
 #include "urcu/map/urcu.h"
 #include "urcu/static/urcu.h"
  */
 #define RCU_QS_ACTIVE_ATTEMPTS 100
 
-/*
- * RCU_MEMBARRIER is only possibly available on Linux.
- */
-#if defined(RCU_MEMBARRIER) && defined(__linux__)
-#include <urcu/syscall-compat.h>
-#endif
-
-/* If the headers do not support SYS_membarrier, fall back on RCU_MB */
-#ifdef SYS_membarrier
-# define membarrier(...)               syscall(SYS_membarrier, __VA_ARGS__)
+/* If the headers do not support membarrier system call, fall back on RCU_MB */
+#ifdef __NR_membarrier
+# define membarrier(...)               syscall(__NR_membarrier, __VA_ARGS__)
 #else
 # define membarrier(...)               -ENOSYS
 #endif
 
-#define MEMBARRIER_EXPEDITED           (1 << 0)
-#define MEMBARRIER_DELAYED             (1 << 1)
-#define MEMBARRIER_QUERY               (1 << 16)
+enum membarrier_cmd {
+       MEMBARRIER_CMD_QUERY = 0,
+       MEMBARRIER_CMD_SHARED = (1 << 0),
+};
 
 #ifdef RCU_MEMBARRIER
 static int init_done;
@@ -100,7 +96,21 @@ void __attribute__((constructor)) rcu_init(void);
 void __attribute__((destructor)) rcu_exit(void);
 #endif
 
+/*
+ * rcu_gp_lock ensures mutual exclusion between threads calling
+ * synchronize_rcu().
+ */
 static pthread_mutex_t rcu_gp_lock = PTHREAD_MUTEX_INITIALIZER;
+/*
+ * rcu_registry_lock ensures mutual exclusion between threads
+ * registering and unregistering themselves to/from the registry, and
+ * with threads reading that registry from synchronize_rcu(). However,
+ * this lock is not held all the way through the completion of awaiting
+ * for the grace period. It is sporadically released between iterations
+ * on the registry.
+ * rcu_registry_lock may nest inside rcu_gp_lock.
+ */
+static pthread_mutex_t rcu_registry_lock = PTHREAD_MUTEX_INITIALIZER;
 struct rcu_gp rcu_gp = { .ctr = RCU_GP_COUNT };
 
 /*
@@ -134,7 +144,7 @@ static void mutex_lock(pthread_mutex_t *mutex)
                        _CMM_STORE_SHARED(URCU_TLS(rcu_reader).need_mb, 0);
                        cmm_smp_mb();
                }
-               poll(NULL,0,10);
+               (void) poll(NULL, 0, 10);
        }
 #endif /* #else #ifndef DISTRUST_SIGNALS_EXTREME */
 }
@@ -149,17 +159,17 @@ static void mutex_unlock(pthread_mutex_t *mutex)
 }
 
 #ifdef RCU_MEMBARRIER
-static void smp_mb_master(int group)
+static void smp_mb_master(void)
 {
        if (caa_likely(rcu_has_sys_membarrier))
-               (void) membarrier(MEMBARRIER_EXPEDITED);
+               (void) membarrier(MEMBARRIER_CMD_SHARED, 0);
        else
                cmm_smp_mb();
 }
 #endif
 
 #ifdef RCU_MB
-static void smp_mb_master(int group)
+static void smp_mb_master(void)
 {
        cmm_smp_mb();
 }
@@ -202,13 +212,13 @@ static void force_mb_all_readers(void)
        cds_list_for_each_entry(index, &registry, node) {
                while (CMM_LOAD_SHARED(index->need_mb)) {
                        pthread_kill(index->tid, SIGRCU);
-                       poll(NULL, 0, 1);
+                       (void) poll(NULL, 0, 1);
                }
        }
        cmm_smp_mb();   /* read ->need_mb before ending the barrier */
 }
 
-static void smp_mb_master(int group)
+static void smp_mb_master(void)
 {
        force_mb_all_readers();
 }
@@ -220,12 +230,29 @@ static void smp_mb_master(int group)
 static void wait_gp(void)
 {
        /* Read reader_gp before read futex */
-       smp_mb_master(RCU_MB_GROUP);
-       if (uatomic_read(&rcu_gp.futex) == -1)
-               futex_async(&rcu_gp.futex, FUTEX_WAIT, -1,
-                     NULL, NULL, 0);
+       smp_mb_master();
+       if (uatomic_read(&rcu_gp.futex) != -1)
+               return;
+       while (futex_async(&rcu_gp.futex, FUTEX_WAIT, -1,
+                       NULL, NULL, 0)) {
+               switch (errno) {
+               case EWOULDBLOCK:
+                       /* Value already changed. */
+                       return;
+               case EINTR:
+                       /* Retry if interrupted by signal. */
+                       break;  /* Get out of switch. */
+               default:
+                       /* Unexpected error. */
+                       urcu_die(errno);
+               }
+       }
 }
 
+/*
+ * Always called with rcu_registry lock held. Releases this lock between
+ * iterations and grabs it again. Holds the lock when it returns.
+ */
 static void wait_for_readers(struct cds_list_head *input_readers,
                        struct cds_list_head *cur_snap_readers,
                        struct cds_list_head *qsreaders)
@@ -247,7 +274,7 @@ static void wait_for_readers(struct cds_list_head *input_readers,
                if (wait_loops >= RCU_QS_ACTIVE_ATTEMPTS) {
                        uatomic_dec(&rcu_gp.futex);
                        /* Write futex before read reader_gp */
-                       smp_mb_master(RCU_MB_GROUP);
+                       smp_mb_master();
                }
 
                cds_list_for_each_entry_safe(index, tmp, input_readers, node) {
@@ -277,15 +304,19 @@ static void wait_for_readers(struct cds_list_head *input_readers,
                if (cds_list_empty(input_readers)) {
                        if (wait_loops >= RCU_QS_ACTIVE_ATTEMPTS) {
                                /* Read reader_gp before write futex */
-                               smp_mb_master(RCU_MB_GROUP);
+                               smp_mb_master();
                                uatomic_set(&rcu_gp.futex, 0);
                        }
                        break;
                } else {
+                       /* Temporarily unlock the registry lock. */
+                       mutex_unlock(&rcu_registry_lock);
                        if (wait_loops >= RCU_QS_ACTIVE_ATTEMPTS)
                                wait_gp();
                        else
                                caa_cpu_relax();
+                       /* Re-lock the registry lock before the next loop. */
+                       mutex_lock(&rcu_registry_lock);
                }
 #else /* #ifndef HAS_INCOHERENT_CACHES */
                /*
@@ -296,21 +327,25 @@ static void wait_for_readers(struct cds_list_head *input_readers,
                if (cds_list_empty(input_readers)) {
                        if (wait_loops >= RCU_QS_ACTIVE_ATTEMPTS) {
                                /* Read reader_gp before write futex */
-                               smp_mb_master(RCU_MB_GROUP);
+                               smp_mb_master();
                                uatomic_set(&rcu_gp.futex, 0);
                        }
                        break;
                } else {
                        if (wait_gp_loops == KICK_READER_LOOPS) {
-                               smp_mb_master(RCU_MB_GROUP);
+                               smp_mb_master();
                                wait_gp_loops = 0;
                        }
+                       /* Temporarily unlock the registry lock. */
+                       mutex_unlock(&rcu_registry_lock);
                        if (wait_loops >= RCU_QS_ACTIVE_ATTEMPTS) {
                                wait_gp();
                                wait_gp_loops++;
                        } else {
                                caa_cpu_relax();
                        }
+                       /* Re-lock the registry lock before the next loop. */
+                       mutex_lock(&rcu_registry_lock);
                }
 #endif /* #else #ifndef HAS_INCOHERENT_CACHES */
        }
@@ -348,17 +383,23 @@ void synchronize_rcu(void)
         */
        urcu_move_waiters(&waiters, &gp_waiters);
 
+       mutex_lock(&rcu_registry_lock);
+
        if (cds_list_empty(&registry))
                goto out;
 
/* All threads should read qparity before accessing data structure
* where new ptr points to. Must be done within rcu_gp_lock because it
* iterates on reader threads.*/
        /* Write new ptr before changing the qparity */
-       smp_mb_master(RCU_MB_GROUP);
+       smp_mb_master();
 
        /*
         * Wait for readers to observe original parity or be quiescent.
+        * wait_for_readers() can release and grab again rcu_registry_lock
+        * interally.
         */
        wait_for_readers(&registry, &cur_snap_readers, &qsreaders);
 
@@ -399,6 +440,8 @@ void synchronize_rcu(void)
 
        /*
         * Wait for readers to observe new parity or be quiescent.
+        * wait_for_readers() can release and grab again rcu_registry_lock
+        * interally.
         */
        wait_for_readers(&cur_snap_readers, NULL, &qsreaders);
 
@@ -407,11 +450,14 @@ void synchronize_rcu(void)
         */
        cds_list_splice(&qsreaders, &registry);
 
-       /* Finish waiting for reader threads before letting the old ptr being
-        * freed. Must be done within rcu_gp_lock because it iterates on reader
-        * threads. */
-       smp_mb_master(RCU_MB_GROUP);
+       /*
+        * Finish waiting for reader threads before letting the old ptr
+        * being freed. Must be done within rcu_registry_lock because it
+        * iterates on reader threads.
+        */
+       smp_mb_master();
 out:
+       mutex_unlock(&rcu_registry_lock);
        mutex_unlock(&rcu_gp_lock);
 
        /*
@@ -447,27 +493,35 @@ void rcu_register_thread(void)
        assert(URCU_TLS(rcu_reader).need_mb == 0);
        assert(!(URCU_TLS(rcu_reader).ctr & RCU_GP_CTR_NEST_MASK));
 
-       mutex_lock(&rcu_gp_lock);
+       mutex_lock(&rcu_registry_lock);
+       assert(!URCU_TLS(rcu_reader).registered);
+       URCU_TLS(rcu_reader).registered = 1;
        rcu_init();     /* In case gcc does not support constructor attribute */
        cds_list_add(&URCU_TLS(rcu_reader).node, &registry);
-       mutex_unlock(&rcu_gp_lock);
+       mutex_unlock(&rcu_registry_lock);
 }
 
 void rcu_unregister_thread(void)
 {
-       mutex_lock(&rcu_gp_lock);
+       mutex_lock(&rcu_registry_lock);
+       assert(URCU_TLS(rcu_reader).registered);
+       URCU_TLS(rcu_reader).registered = 0;
        cds_list_del(&URCU_TLS(rcu_reader).node);
-       mutex_unlock(&rcu_gp_lock);
+       mutex_unlock(&rcu_registry_lock);
 }
 
 #ifdef RCU_MEMBARRIER
 void rcu_init(void)
 {
+       int ret;
+
        if (init_done)
                return;
        init_done = 1;
-       if (!membarrier(MEMBARRIER_EXPEDITED | MEMBARRIER_QUERY))
+       ret = membarrier(MEMBARRIER_CMD_QUERY, 0);
+       if (ret >= 0 && (ret & MEMBARRIER_CMD_SHARED)) {
                rcu_has_sys_membarrier = 1;
+       }
 }
 #endif
 
@@ -488,9 +542,9 @@ static void sigrcu_handler(int signo, siginfo_t *siginfo, void *context)
  * rcu_init constructor. Called when the library is linked, but also when
  * reader threads are calling rcu_register_thread().
  * Should only be called by a single thread at a given time. This is ensured by
* holing the rcu_gp_lock from rcu_register_thread() or by running at library
* load time, which should not be executed by multiple threads nor concurrently
* with rcu_register_thread() anyway.
  */
 void rcu_init(void)
 {
@@ -348,17 +383,23 @@ void synchronize_rcu(void)
                goto out;
 
-       /* All threads should read qparity before accessing data structure
-        * where new ptr points to. Must be done within rcu_gp_lock because it
-        * iterates on reader threads.*/
+       /*
+        * All threads should read qparity before accessing data structure
+        * where new ptr points to. Must be done within rcu_registry_lock
+        * because it iterates on reader threads.
+        */
        /* Write new ptr before changing the qparity */
@@ -488,9 +542,9 @@ static void sigrcu_handler(int signo, siginfo_t *siginfo, void *context)
  * reader threads are calling rcu_register_thread().
  * Should only be called by a single thread at a given time. This is ensured by
- * holing the rcu_gp_lock from rcu_register_thread() or by running at library
- * load time, which should not be executed by multiple threads nor concurrently
- * with rcu_register_thread() anyway.
+ * holing the rcu_registry_lock from rcu_register_thread() or by running
+ * at library load time, which should not be executed by multiple
+ * threads nor concurrently with rcu_register_thread() anyway.
  */
 void rcu_init(void)
This page took 0.026422 seconds and 4 git commands to generate.